
import numpy as np
from matplotlib import pyplot as plt
import numpy.random as rd 

### Ajustement affine, lorque le modèle physique est du type Y = aX + b

def AjustAff_MC(Lx,Ly,sigma,fig = True):
    """ ajustement affine sans pondération d'une nuage de points expérimentaux

    Parameters
    ----------
    Lx : tableau 1D np.array([...,...])
        les valeurs des abscisses xi
    Ly : tableau 1D np.array([...,...])
        les valeurs des abscisses yi correspondant aux xi
    sigma : nombre flottant
        l'incertitude-type sur les yi (supposée la même pour tout i)
        si sigma == 'inconnu', l'incertitude-type sur les yi est déduite des écarts à la droite d'ajustement
    fig : booléen
        par défaut fig = True et une figure représentant le nuage de points expérimentaux et la droite d'ajustement affine est tracée, ainsi qu'une figure représentant les résidus normalisés
        
    Print
    ----------
    valeurs de la pente a, de u(a), de l'ordonnée à l'origine b et de u(b)
    
    Returns
    -------
    a_MC, b_MC : np.array([...,...]), np.array([...,...])
        les tableaux des pentes et des ordonnées à l'origine 
    """
    p = np.polyfit(Lx, Ly, 1) # fonction python pour effectuer un ajustement affine (ajustement polynomial de degré 1)
    a,b = p[0],p[1]   # valeurs des coefficients basés sur la réalisation expérimentale
    print(a,b)

    def ajustement(x):
        return a*x+b
    
    Ly_ajuste = ajustement(Lx)
    N_MC = 100000 # nombre d'expériences simulées, pas le nombre de mesures...
    a_MC, b_MC = np.zeros(N_MC),np.zeros(N_MC)   # initialisation des listes dans lesquelles on va stocker les pentes et ordonnées à l'origine simulées
 
    for i in range(N_MC):  # pour chaque expérience
    # on simule la mesure de Y pour toutes les valeurs de X
        # Ly_MC = Ly_ajuste + rd.normal(0, sigma, size = len(Lx))
        Ly_MC = Ly_ajuste + rd.uniform(-sigma*3**0.5,sigma*3**0.5, size = len(Lx))
    # on reprend l'ajustement affine avec cette série de valeurs de n
        p = np.polyfit(Lx, Ly_MC, 1)
    # on stocke les valeurs des paramètres d'ajustement dans les listes AMC et BMC
        a_MC[i] = p[0]
        b_MC[i] = p[1]
        
    a_moy = np.mean(a_MC)
    u_a = np.std(a_MC, ddof = 1)
    print('paramètre a : ',a_moy)
    
    print('incertitude-type : u(a) = ',u_a)
    
    b_moy = np.mean(b_MC)
    u_b = np.std(b_MC, ddof = 1)
    print('paramètre b : ',b_moy)
    print('incertitude-type : u(b) = ', u_b)
        
    if fig == True:
        X = [Lx[0],Lx[-1]]
        Y = [a_moy*Lx[0]+b_moy,a_moy*Lx[-1]+b_moy]
        plt.figure('points et ajustement affine')
        plt.plot(Lx,Ly, 'o', label = 'points expérimentaux',color = 'red')
        plt.errorbar(Lx, Ly, yerr = sigma, fmt = 'm.',ecolor = 'black',elinewidth = 3)
        plt.xlabel('$X$')
        plt.ylabel('$Y$')
        plt.plot(X,Y, "r-", label = 'ajustement affine',color = 'b',alpha = 0.5)
        plt.legend(fontsize = 15)
        plt.show()

        Z = (Ly-(a_moy*Lx+b_moy))/(sigma)
        plt.figure('résidus normalisés ajustement affine Monte Carlo')
        plt.plot(Lx, Z,'o',color = 'red')
        plt.fill_between([np.min(Lx),np.max(Lx)], y1 = -2, y2 = 2, color = 'olive',alpha = 0.4)
        plt.xlim(.9*np.min(Lx), 1.02*np.max(Lx))
        plt.xlabel('$X$')
        plt.ylabel('résidus normalisés de $Y$' ), plt.ylim(-3,3)
        plt.ticklabel_format(axis = 'x', style = 'sci', scilimits = (0,0))
        plt.grid()
        plt.show()
    
    return (a_MC,b_MC)
    
### Ajustement linéaire, lorque le modèle physique est du type Y = aX

def linfit(Lx,Ly): # calcul de la pente a de la droite de régression linéaire à parir des données expérimentales
    N = len(Lx)
    xy, x2 = 0,0
    for i in range(N):
        xy = xy + Lx[i]*Ly[i]
        x2 = x2 + Lx[i]**2
    return xy/x2
    
def AjustLin_MC(Lx,Ly,sigma,fig = True):
    """ ajustement affine sans pondération d'une nuage de points expérimentaux

    Parameters
    ----------
    Lx : tableau 1D np.array([...,...])
        les valeurs des abscisses xi
    Ly : tableau 1D np.array([...,...])
        les valeurs des abscisses yi correspondant aux xi
    sigma : nombre flottant
        l'incertitude-type sur les yi (supposée la même pour tout i)
        si sigma == 'inconnu', l'incertitude-type sur les yi est déduite des écarts à la droite d'ajustement
    fig : booléen
        par défaut fig = True et une figure représentant le nuage de points expérimentaux et la droite d'ajustement affine est tracée, ainsi qu'une figure représentant les résidus normalisés
        
    Print
    ----------
    valeurs de la pente a, de u(a)
    
    Returns
    -------
    a_MC: np.array([...,...])
        le tableaux des pentes 
    """
    a = linfit(Lx, Ly) # fonction python pour effectuer un ajustement linéaire
    print(a)
    

    def ajustement(x):
        return a*x
    
    Ly_ajuste = ajustement(Lx)
    N_MC = 100000 # nombre d'expériences simulées, pas le nombre de mesures...
    a_MC = np.zeros(N_MC)   # initialisation des listes dans lesquelles on va stocker les pentes simulées
 
    for i in range(N_MC):  # pour chaque expérience
    # on simule la mesure de Y pour toutes les valeurs de X
        # Ly_MC = Ly_ajuste + rd.normal(0, sigma, size = len(Lx))
        Ly_MC = Ly_ajuste + rd.uniform(-sigma*3**0.5,sigma*3**0.5, size = len(Lx))
    # on reprend l'ajustement affine avec cette série de valeurs de n
        a_MC[i] = linfit(Lx, Ly_MC)
    # on stocke les valeurs des paramètres d'ajustement dans la liste a_MC
    a_moy = np.mean(a_MC)
    u_a = np.std(a_MC, ddof = 1)
    print('paramètre a : ',a_moy)
    print('incertitude-type : u(a) = ',u_a)
    
    if fig == True:
        X = [Lx[0],Lx[-1]]
        Y = [a_moy*Lx[0],a_moy*Lx[-1]]
        plt.figure('points et ajustement linéaire')
        plt.plot(Lx,Ly, 'o', label = 'points expérimentaux',color = 'red')
        plt.errorbar(Lx, Ly, yerr = sigma, fmt = 'm.',ecolor = 'black',elinewidth = 3)
        plt.xlabel('$X$')
        plt.ylabel('$Y$')
        plt.plot(X,Y, "r-", label = 'ajustement lineaire',color = 'b',alpha = 0.6)
        plt.legend(fontsize = 15)
        plt.show()

        Z = (Ly-(a_moy*Lx))/(sigma)
        plt.figure('résidus normalisés ajustement linéaire Monte Carlo')
        plt.plot(Lx, Z,'o',color = 'red')
        plt.fill_between([np.min(Lx),np.max(Lx)], y1 = -2, y2 = 2, color = 'olive', alpha = .25)
        plt.xlim(.9*np.min(Lx), 1.02*np.max(Lx))
        plt.xlabel('$X$')
        plt.ylabel('résidus normalisés de $Y$' ), plt.ylim(-3,3)
        plt.ticklabel_format(axis = 'x', style = 'sci', scilimits = (0,0))
        plt.grid()
        plt.show()
    
    return (a_MC)
    
    

## Exemple

Lx = np.array([0,1,2,3,4,5,6,7,8,9]) # les abscisses, incertitudes négligées
Ly = 10*Lx + rd.uniform(-2*3**0.5,2*3**0.5, size = len(Lx)) # les ordonnées, incertitudes-type identiques égales à 2 avec une loi uniforme

sigma = 2

a_MC,b_MC = AjustAff_MC(Lx,Ly,sigma) # un ajustement affine donne a = 10,18 avec u(a) = 0,22 et b = -0,59 avec u(b) = 1,2

a_MC = AjustLin_MC(Lx,Ly,sigma) # un ajustement linéaire donne a = 10,09 avec u(a) = 0,11. Il est ici mieux adapté car la loi physique sous-jacente est linéaire, pas affine. Une recherche de loi affine amène à sur-estimer l'incertitude-type sur a.